#ifndef _CHRONOPTICS_TOF_BASE_HPP_
#define _CHRONOPTICS_TOF_BASE_HPP_

#include <chronoptics/tof/error.h>

#include <array>
#include <functional>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <vector>

namespace chronoptics {
namespace tof {


namespace detail {

struct error {
  error() = default;
  ~error() {
    if (ptr_) {
      tof_error_delete(ptr_);
    }
  }

  tof_error_t ptr_{nullptr};
};

class ThrowOnError {
 public:
  ~ThrowOnError() noexcept(false) {
    if (_error.ptr_) {
      throw std::runtime_error(tof_error_message(_error.ptr_));
    }
  }

  /** This sets the error pointer in the c interface */
  operator tof_error_t *() { return &_error.ptr_; }

 private:
  error _error;
};

}

#ifndef TOF_ERROR_HANDLER
#define TOF_ERROR_HANDLER detail::ThrowOnError
#endif

namespace detail {

template <typename Ret, typename... Args,
          typename std::enable_if<std::is_void<Ret>::value, bool>::type = true>
Ret callback_handler(Args... args, void *user_data) {
  auto *func = reinterpret_cast<std::function<Ret(Args...)> *>(user_data);
  try {
    return (*func)(args...);
  } catch (...) {
    return;
  }
}

template <typename Ret, typename... Args,
          typename std::enable_if<!std::is_void<Ret>::value, bool>::type = true>
Ret callback_handler(Args... args, void *user_data) {
  auto *func = reinterpret_cast<std::function<Ret(Args...)> *>(user_data);
  try {
    return (*func)(args...);
  } catch (...) {
    return Ret{};
  }
}

template <typename ClassPtr, void (*Destructor)(ClassPtr*)>
class Base {
 public:
  Base() = default;
  ~Base() {
    if (ptr_) Destructor(ptr_);
  }

  Base(const Base&) = delete;
  Base& operator=(const Base&) = delete;

  Base(Base&& old) {
    ptr_ = old.ptr_;
    old.ptr_ = nullptr;
  }

  Base& operator=(Base&& old) {
    if (ptr_) Destructor(ptr_);

    ptr_ = old.ptr_;
    old.ptr_ = nullptr;
    return *this;
  }

  operator bool() const { return ptr_ != nullptr; }

 protected:
  ClassPtr *ptr_{nullptr};
};

}

class StringView {
 public:
  StringView(const std::string &string) : string_(string.c_str()) {}
  StringView(const char *string) : string_(string) {}

  operator const char *() const { return string_; }

 private:
  const char *string_;
};

} // tof
} // chronoptics

#endif
